#!/Perl\bin\MSWin32-x86\Perl 

# This program reads through a *_Bodies.csv file created by the AS CSV exporter and assigns a weight to each system.
# It will generate an economic value to each system based on the presence of white dwarf stars, gas giants, high density planets,
# Asteroid belts, and the population of the system.
#
# Adds routes between stars with most routes being less than 8LY.  However, longer routes are added so that all stars are reachable
# via routes.  The minimum number of the shortest routes are added.
#
# An IP pro file with all of the systems in the sector is created with systems sorted by Major, Minor and Offroute systems.  Each 
# entry in the table is weighted with its total economic score, that is the inherent economic value of the system, plus the value 
# of trade and travel through the system between the major systems.  Minor systems are those that are on a route between Major 
# systems.  Offroute systems are all other systems.
#
# Finally, a csv file is created with the route information that can be imported into AS.
#
# The script can be stopped at any point and it will pick up at the last fully completed stage.
#
# All of this is done in the same directory the program is run from.
# Version 1.1
# Fixed initialization bug in line 158, added traffic calculation in lines 842-850.  Sigh.

use strict;
use warnings;
use Text::ParseWords;
use Graph;
use Graph::Undirected;

# System Variables
my $LYLIMIT = 8;		# The maximum normal trade route distance.  Set this to something suitable for your sector.
						# It will be exceeded the minimum number of times in order to link all systems to all other systems.

# Global Variables
my $MAXWEIGHT = 0;		# The graph program uses weight as a penalty, so a low value is better, start out at 0.
my $TRAFFICFACTOR = 1;
my $filename = "";		# The filename should be passed into the program
my @systems;			# The array of nodes from CollatedSystemData.csv
my @routes;				# The array of edges from Routes.csv
my @newroutes;			# The array of newly created edges.
my %IDpointer;			# The hash of SystemID to @systems index to make finding specific systems faster.
my $cluster = 1;		# The current cluster number, 0 indicates that no cluster analysis has been done.
my %SystemWeights;		# A mapping of ID to weight.
my $space = Graph::Undirected->new;
my %visited;			# A list of all visited nodes during the search.
my %NameToID;			# A mapping of name to ID for easy reference.
my @ingraph;

#----------------------------------------------------------------------------------------------------------------------
sub parse_csv0 {
	return quotewords(",", 0, $_[0]);
}
#----------------------------------------------------------------------------------------------------------------------
sub calc_value {

	my $filename = $_[0];
	my $dataline;
	my @datalinearray;
	my $gasgiants = 0;
	my $score = 0;
	my $habitability = 0;
	my $pop = 0;
	my $WDpresent = 0;
	my $thisbodytype = 0;
	my $thisdensity = 0;
	my $thispopulation = -1;
	my $thishabitability = -1;
	my $k = -1;

	# Open Data file.
	# 0 BodyID
	# 1 BodyName	
	# 2 BodyStringID	
	# 3 ParentID			Will be blank for the system
	# 4 Parent	
	# 5 ParentStringID	
	# 6 Albedo	
	# 7 AngleAscendingNode	
	# 8 AnglePeriapsis	
	# 9 Atmosphere	
	# 10 AtmosphereNotes	
	# 11 AxialTilt	
	# 12 BlazonDisplay	
	# 13 BlazonImageFile	
	# 14 BlazonWrap	
	# 15 BodyCount	
	# 16 BodyType	
	# 17 ChildHabitability	
	# 18 ChildPopulation	
	# 19 Color	
	# 20 Density	
	# 21 Distance	
	# 22 DSOIntensity	
	# 23 Eccentricity	
	# 24 FontStyle	
	# 25 GMNotes	
	# 26 Habitability	
	# 27 Inclination	
	# 28 LabelColor	
	# 29 LabelDistance	
	# 30 Luminosity	
	# 31 Mass	
	# 32 Notes	
	# 33 PoliticalAffiliation	
	# 34 Population	
	# 35 PreviewImageFile	
	# 36 Radius	
	# 37 RandomSeed	
	# 38 RetrogradeOrbit	
	# 39 Rotation	
	# 40 RouteDistance	
	# 41 Spectral	
	# 42 SphereInfluence	
	# 43 SphereInfluenceColor	
	# 44 SphereInfluenceSize	
	# 45 Temp	
	# 46 TimeOffset	
	# 47 TypeID	
	# 48 Visible	
	# 49 Water	
	# 50 X	
	# 51 Y	
	# 52 Z

	open (DATAFILE, "$filename") ||
		die "Can't open $filename datafile:$!";

	# If the $thisparentid value is blank then the body is a root object and will get a net score.
	$dataline = <DATAFILE>; # Skip the first line of the file.
	while (defined ($dataline = <DATAFILE>)) {
		chomp ($dataline);
		@datalinearray = parse_csv0 ($dataline);
		$thisbodytype = $datalinearray[47];
		$thisdensity = $datalinearray[20];
		$thispopulation = $datalinearray[34];
		$thishabitability = $datalinearray[26];

		# When a line with no parentid is found we have finished the previous system.  But we forget the final system.
		if ($datalinearray[3] eq "") {
			# If this is not the first time through the file, Close out the previous system.
			if ($k > -1) {
				# If there are no gas giants, - 50 points
				if ($gasgiants < 1) {
					$score -= 50;
				}
				$systems[$k][5] = $score;			# Economic score
				if ($pop >= 10000000) {
					$systems[$k][6] = 1;			# Major System 1=Major
				}
				$systems[$k][7] = $habitability;	# Habitability 1=Habitable, 2=Hospitable
				$systems[$k][8] = $WDpresent;		# WD present? 1=White Dwarf in system
				if ($WDpresent > 0) {				# Having a WD present simply destroys any value to the system.
					$systems[$k][5] = -200;
				}
				$systems[$k][10] = $pop;
			}
			# Go on to the next system.
			$k++;
			$systems[$k] = [($datalinearray[0],$datalinearray[1],$datalinearray[50],$datalinearray[51],$datalinearray[52],0,0,0,0,0,0)];
			$IDpointer{$systems[$k][0]} = $k;

			# We also have to reset all of the carried values.
			# Reset the system wide values.
			$score = 0;
			$gasgiants = 0;
			$habitability = 0;
			$pop = 0;
			$WDpresent = 0;

		} else {
			# For each child do this.
			if ($thisbodytype == 8) { #This is a white dwarf
				$WDpresent = 1;
			} else {
				# 4 = Space Station
				# 100 = terrestrial
				# 102 = Astroid belt
				# 101 = Gas Giant
				# 104 = Ring
				# 103 = Planetoid/Small body

				# Count up all gas giants.
				if ($thisbodytype == 101) { #Gas Giant
					$gasgiants++;
				}
				# Count up all high density planets.  +5 points each
				if (($thisbodytype == 100) && ($thisdensity >=1.0)) {
					$score += 2;
				}
				# Count up all high density planetoids. +3 points each
				if (($thisbodytype == 103) && ($thisdensity >=1.0)) {
					$score += 1;
				}
				# Count up all asteroid belts.  +10 points each
				if ($thisbodytype == 102) {
					$score += 3;
				}
				# Count up all high population planets (1 000 000).  +100 Points
				if (($thisbodytype != 4) && ($thispopulation >= 1000000)) { # not a space station
					$score += 100;
				}
				# Add the population of the child.
				$pop += $thispopulation;

				if ($thishabitability > $habitability) {
					$habitability = $thishabitability;
				}
			}
		}
	}
	# Close out the final system.  ?
	if ($gasgiants < 1) {
		$score -= 50;
	}
	$systems[$k][5] = $score;			# Economic score
	if ($pop >= 10000000) {
		$systems[$k][6] = 1;			# Major System 1=Major
	}
	$systems[$k][7] = $habitability;	# Habitability 1=Habitable, 2=Hospitable
	$systems[$k][8] = $WDpresent;		# WD present? 1=White Dwarf in system
	if ($WDpresent > 0) {				# Having a WD present simply destroys any value to the system.
		$systems[$k][5] = -200;
	}
	$systems[$k][10] = $pop;
	close (DATAFILE);
}
#----------------------------------------------------------------------------------------------------------------------
sub FindDistance {
	my $x1 = shift @_;
	my $x2 = shift @_;
	my $y1 = shift @_;
	my $y2 = shift @_;
	my $z1 = shift @_;
	my $z2 = shift @_;

	my $thisdist = 0;

	if ( (abs ($x1-$x2) < $LYLIMIT*2) &&
		(abs ($y1-$y2) < $LYLIMIT*2) && 
		(abs ($z1-$z2) < $LYLIMIT*2) ) { # first check if the star is near your starting point.
		$thisdist = ($x1-$x2)**2;
		$thisdist = $thisdist + ($y1-$y2)**2; 
		$thisdist = $thisdist + ($z1-$z2)**2;
		$thisdist = sqrt($thisdist);
		if ($thisdist < $LYLIMIT) {
			$thisdist = sprintf ("%.2f", $thisdist);
			return $thisdist;
		}
	}
	return -1;
}
#----------------------------------------------------------------------------------------------------------------------
sub traverse_cluster { #takes a SystemID as the input, is recursive
	my $thissystemID = shift @_;
	# Recursion
	#10 Set SystemID[9] = cluster
	#20 Find all edges that have SystemID in [1] or [2]
	for (my $s=0; $s<=$#routes; $s++) {
		if ($routes[$s][1] == $thissystemID) {
			#30 Push the OtherID to a stack, if it has not already been visited
			if ($systems[$IDpointer{$routes[$s][2]}][9] < 1) {
				$systems[$IDpointer{$routes[$s][2]}][9] = $cluster;
				push @_, $systems[$IDpointer{$routes[$s][2]}][0];
			}
		} elsif ($routes[$s][2] == $thissystemID) {
			#30 Push the OtherID to a stack, if it has not already been visited
			if ($systems[$IDpointer{$routes[$s][1]}][9] < 1) {
				$systems[$IDpointer{$routes[$s][1]}][9] = $cluster;
				push @_, $systems[$IDpointer{$routes[$s][1]}][0];
			}
		}
	}
	#40 if the stack is not empty Pop an OtherID from the stack, go to 10 with OtherID
	if ($#_ >= 0) {
		&traverse_cluster;
	} #50 if the stack is empty go to 60
	return;
}
#----------------------------------------------------------------------------------------------------------------------
sub FindAnyDistance {
	my $x1 = shift @_;
	my $x2 = shift @_;
	my $y1 = shift @_;
	my $y2 = shift @_;
	my $z1 = shift @_;
	my $z2 = shift @_;

	my $thisdist = 0;

	$thisdist = ($x1-$x2)**2;
	$thisdist = $thisdist + ($y1-$y2)**2; 
	$thisdist = $thisdist + ($z1-$z2)**2;
	$thisdist = sqrt($thisdist);
	$thisdist = sprintf ("%.2f", $thisdist);
	return $thisdist;
}
#----------------------------------------------------------------------------------------------------------------------
sub breadth2search {

	my $graph = shift;
	my $source = shift;
	my $destination = shift;
	my @route;
	my @q;
	my %walker;
	my $thisweight = 0;

	# Verify that the source and destination node exist in the graph otherwise quit
	if ( !( $graph->has_vertex( $source ) ) || !($graph->has_vertex( $destination ) ) ) {
		print "Source $source or Destination $destination are not in the graph\n";
		print "Walker is: %visited\n";
		return -1;
	} 

	# Add the source to the queue.
	push @q, $source;
	$walker { $source } = ["", 0]; # predecessor, Jumps, cost

	# While there are vertices in the queue.
	while (defined ( my $v = shift @q ) ) { # $v is this vertex.
		if ( !defined( $visited { $v } ) ) {

			my $currentdistance = $walker{ $v }->[1];

			# Mark $v as visited.
			$visited{ $v } = 1;

			# Found it.
			if ($v eq $destination) {
				# walk back.
				my @route;
				my $totaldistance = $walker{ $v }->[1];
				my $n;		 # $n is index node name.

				while ($v ne $source) { # while we are not at home.
					# push me onto the stack
					push @route, $v ;

					# Look at predecessor
					$v = $walker{ $v }->[0];
				}
				push @route, $source;
				return @route;
			}

			# Get a list of all edges from $v and append them.
			my @e = $graph->neighbors( $v );

			# Debug
			#print "Num Edges: $#e Edges: @e\n";
			#/ Debug

			foreach my $u (@e) { # $u is a vertex index for this foreach loop.
				if ( !defined( $visited{ $u } ) ) {

					$thisweight = $graph->get_vertex_attribute( $u, 'weight' );

					# Debug
					#print "cost= $thiscost.\n";
					#/ Debug

					if ( !defined( $walker{ $u } ) ) {
						$walker{ $u } = [$v, $currentdistance + $thisweight];
					}
					elsif ( $walker{ $u }->[1] > $currentdistance + $thisweight ) {
						$walker{ $u }->[0] = $v;
						$walker{ $u }->[1] = $currentdistance + $thisweight;
					}
					push @q, $u;
				}
			}
			my @tmp = sort { $walker{$a}->[1] <=> $walker{$b}->[1] } @q;
			@q = @tmp; 
		}
	}
}

#-----------------------------------------------------------------------------------

sub isingraph {
	my $testnode = shift;
	
	foreach my $n  (@ingraph) {
		if ($n eq $testnode) {
			return 1;
		}
	}
	return 0;
}

#-----------------------------------------------------------------------------------

sub SaveEdge {

# If the traffic attribute is > 0 then the edge should be printed.

	my $source = shift;
	my $target = shift;
	my $thisedgelabel = $space->get_edge_attribute ($source, $target, "label");
	my $traffic = $space->get_edge_attribute ($source, $target, "traffic");
	$traffic = $traffic + 2;
	$space->set_edge_attribute ($source, $target, "traffic", $traffic);

}

#-----------------------------------------------------------------------------------
# MAIN PROGRAM

# CollatedSystemData.csv data format:
# 0 = ID
# 1 = System Name
# 2 = X coordinate
# 3 = Y coordinate
# 4 = Z coordinate
# 5 = Weight
# 6 = Major?  1=Major System
# 7 = Habitability 1=Habitable 2=Hospitable
# 8 = WD Present? 1=White Dwarf in system
# 9 = First cluster ID, then Econ value, sum of all traffic from all edges that connect
# 10 = Total system population.

# Routes.csv format
# 0 = route ID
# 1 = Start ID. A system ID
# 2 = End ID. A system ID
# 3 = Distance
# 4 = Weight.  Starts out at 0, will be updated when we determine the MSRs

my $thisdist;
my $routeID = 0;
my $SystemID = 1;	# The current systemID that is being traversed.
my $thistime = 0;
my $lasttime = 0;
my $endtime = 0;	# The time the script ended.
my $starttime = time;	# The time the script started.
my $Newroute = 0;	# The total number of new routes found.
my $startingclusterID = 0;	# The cluster ID that is being merged.
my $clusterID = 0;	# The current cluster ID.
my $newclusternum = 999999;
my @exitroute = (0,0,0,100,0);
my $source = 0;
my $destination = 0;
my $traffic = 0;
my $thisscore = 0;
my $found = 0;
my $testfilename = 1;
my $sectorname = "";
my $outputfilename = "";
my $starti = -1;
my $dataline;
my @datalinearray;


#******************************************************************************************************
#******************************************************************************************************

while ($testfilename) {
	print "What is the name of the sector to process?\n";
	$sectorname = <STDIN>;
	chomp $sectorname;
	$filename = $sectorname . "_Bodies.csv";

	# Test filename for existence.
	$testfilename = 0;
	-e $filename or $testfilename = 1;
	if ($testfilename) {
		print "Unable to find $filename:$!\n";
	}
}

# Check for filename_CollatedData.csv
$filename = $sectorname . "_CollatedData.csv";
if (-e $filename) {
	# If it exists, look for the first line for the index to start the MSR calculation.
	# Load CollatedData
	open (OLDDATA, "$filename") ||
		die "Unable to restart from where we left off: $!\nIf you want to start over, remove $filename and $sectorname _Routes.csv.\n";
	my $k = 0;
	# If restarting, the first line will be the initial value for the MSR calculation
	$dataline = <OLDDATA>;
	chomp ($dataline);
	$starti = $dataline +1;
	while (defined ($dataline = <OLDDATA>)) {
		chomp ($dataline);
		@datalinearray = parse_csv0 ($dataline);
		$systems[$k] = [@datalinearray];
		$IDpointer{$datalinearray[0]} = $k;
		$k++;
	}
	close (OLDDATA);

	for (my $i = 0; $i < $#systems; $i++) {
		if ($systems[$i][5] > $MAXWEIGHT) {
			$MAXWEIGHT = $systems[$i][5];
		}
	}

	# Load Routes
	$filename = $sectorname . "_Routes.csv";
	$k = 0;
	open (OLDDATA, "$filename") ||
		die "Unable to restart from where we left off: $!\nIf you want to start over, remove $filename and $sectorname _CollatedData.csv.\n";
	while (defined ($dataline = <OLDDATA>)) {
		chomp ($dataline);
		@datalinearray = parse_csv0 ($dataline);
		$routes[$k++] = [@datalinearray];
	}
	close (OLDDATA);
	# jump down to Stage 5, the MSR calculation

} else {
	# Calculate and write the weights for each system.
	$outputfilename = $sectorname . "_CollatedData.csv";
	print "Stage 1:Determining System Economic Values ...\n";
	$filename = $sectorname . "_Bodies.csv";
	calc_value ($filename);

	for (my $i = 0; $i < $#systems; $i++) {
		if ($systems[$i][5] > $MAXWEIGHT) {
			$MAXWEIGHT = $systems[$i][5];
		}
	}

	$thistime = sprintf ("%.3f", (time - $starttime)/60);
	$lasttime = time;
	print "\tdone ($thistime min).\n";  

	#******************************************************************************************************
	#******************************************************************************************************
	# Create the routes
	print "Stage 2:Creating star routes...\n";
	for (my $i = 0; $i < $#systems; $i++) {
		for (my $k = $i+1; $k < $#systems; $k++) {
			$thisdist = FindDistance($systems[$i][2], $systems[$k][2], $systems[$i][3], $systems[$k][3], $systems[$i][4], $systems[$k][4]);
			if ($thisdist > -0.1) {
				$routes[$routeID] = [($routeID,$systems[$i][0],$systems[$k][0],$thisdist,0)];
				$routeID++;
			}
		}
	}
	print "\tFound $routeID links.\n";
	$thistime = sprintf ("%.3f", (time - $lasttime)/60);
	$lasttime = time;
	print "\tdone ($thistime min).\n";  

	#******************************************************************************************************
	#******************************************************************************************************
	# Assign cluster values
	print "Stage 3:Starting star route cluster analysis...\n";
	print "\t(You may see one or more 'Deep recursion on subroutine main::traverse_cluster' messages, this is normal.)\n";
	$SystemID = 1;
	$systems[$IDpointer{$SystemID}][9] = $cluster;
	traverse_cluster ($SystemID);

	#60 For i=0 to $#systems
	for (my $thisnode=0; $thisnode <= $#systems; $thisnode++) {
		#70 If systems[i][9] equals 0 set cluster+1, go to 10
		if ($systems[$thisnode][9] == 0) {
			$cluster++;
			$SystemID = $systems[$thisnode][0];
			$systems[$IDpointer{$SystemID}][9] = $cluster;
			traverse_cluster($SystemID);
		}
	}
	print "\tFound $cluster clusters.\n";

	$thistime = sprintf ("%.3f", (time - $lasttime)/60);
	$lasttime = time;
	print "\tdone ($thistime min).\n";  

	#******************************************************************************************************
	#******************************************************************************************************
	# merge clusters
	print "Stage 4:Merging clusters...\n";

	$startingclusterID = $cluster;  # Start with the last cluster
	$routeID = $#routes+1; # One more than the last read routeID.

	# Start with the highest cluster number
	# Count down to cluster ID = 2.  All stars should be in clusterID 1.
	for ($clusterID = $startingclusterID; $clusterID > 1; $clusterID--) {
		# Look at each star that has the same cluster number
		for (my $clustersystem = 0; $clustersystem <= $#systems; $clustersystem++) {
			if ($systems[$clustersystem][9] == $clusterID) {
				# Look at each star that does not have the same cluster number
				for (my $systemnum = 0; $systemnum < $#systems; $systemnum++) {
					if ($systems[$systemnum][9] != $clusterID) {
						# calculate the distance from that star to each of the stars in the cluster
						$thisdist = FindAnyDistance($systems[$clustersystem][2], $systems[$systemnum][2], $systems[$clustersystem][3], 
							$systems[$systemnum][3], $systems[$clustersystem][4], $systems[$systemnum][4]);
						# If the distance is shorter that the previously shortest distance, save it.
						if ($thisdist < $exitroute[3]) {
							# 0 = route ID
							# 1 = Start ID. A system ID
							# 2 = End ID. A system ID
							# 3 = Distance
							# 4 = Weight.  Starts out at 0, will be updated when we determine the MSRs
							@exitroute = ($routeID,$systems[$clustersystem][0],$systems[$systemnum][0],$thisdist,0);
							$newclusternum = $systems[$systemnum][9];
						}
					}
				}
				# write out the shortest found distance.
				$routes[$exitroute[0]] = [@exitroute];
				# change the cluster number to the one of the star that terminates the shortest route.
				for (my $i = 0; $i <= $#systems; $i++) {
					if ($systems[$i][9] == $clusterID) {
						#	This should be lower than the previous cluster number..
						if ($newclusternum >= $systems[$i][9]) {
							print "ERROR: new cluser number $newclusternum is greater than or equal to old cluster number $systems[$i][9]!\n";
						}
						if ($exitroute[3] <= 4.00) {
							print "ERROR: too short link to new cluster from $systems[$IDpointer{$exitroute[1]}][1] to $systems[$IDpointer{$exitroute[2]}][1] is $exitroute[3] LY.\n";
						}
						$systems[$i][9] = $newclusternum;
					}
				}
			}
		}
		# reset values.
		@exitroute = (0,0,0,100,0);
		$routeID++;
		$Newroute++;
		$newclusternum = 999999;
	}

	print "\tFound $Newroute new routes.\n\tTotal routes $#routes+1 Done!\n";

	$thistime = sprintf ("%.3f", (time - $lasttime)/60);
	print "\tdone ($thistime min).\n";  
}
$lasttime = time;

#******************************************************************************************************
#******************************************************************************************************
# Determine MSRs
print "Stage 5:Finding trade routes....\n";

my $target;
my $label;
my @results;
my $thistraffic = 0;
my @MajorSystems;		# List of IDs of systems with major = 1
my $NumMSRs = 0;
my $continue = "x";

# Build up extra data structures
for (my $s = 0; $s < $#systems; $s++) {
	if ($systems[$s][6]) {
		push @MajorSystems, $systems[$s][1];
	}
	$NameToID{$systems[$s][1]} = $systems[$s][0];

	#Add this data to the Graph
	$space->add_vertex($systems[$s][0]);
	$space->set_vertex_attribute( $systems[$s][0], 'weight', $MAXWEIGHT - $systems[$s][5] );
}
$NumMSRs = $#MajorSystems + 1;
for (my $r = 0; $r < $#routes; $r++) {
	$source = $routes[$r][1];
	$target = $routes[$r][2];

	# Add this data to the Graph
	$space->add_weighted_edge( $source, $target, 1 );
	$space->set_edge_attribute( $source, $target, 'label', $routes[$r][3] );
	$space->set_edge_attribute( $source, $target, 'traffic', $routes[$r][4] );
}

print "Note, this process can take a very long time.  This sector has $NumMSRs Major \n";
$NumMSRs = ($NumMSRs * $NumMSRs)/2 + 1;
print "Systems. My computer can handle about 100 routes in an hour, this sector will \n";
print "have about $NumMSRs routes.  Actual computation time will depend highly on the \n";
print "complexity of the sector, larger sectors will take longer to generate each route.\n";
print "\nThe following are major systems in this sector:\n";
foreach my $sysname (@MajorSystems) {
	print "\t$sysname\n";
}

# Ask the user if they would like to stop to add more Major Systems.
print "\nIf you would like to change the list of major systems, stop the program now and \n";
print "edit field 7 of CollatedSystemsData.csv.  A '1' will indicate a major system.  Do \n";
print "not use a spreadsheet program to edit the file, it will add quotation marks that \n";
print "will prevent this program from loading the data set.\n\n";
print "Would you like to continue? (y/n/yes/no) ";
while ($continue !~ m/Y|N|y|n|Yes|No|yes|no|YES|NO/ ) {
	$continue = <STDIN>;
	chomp $continue;
	if ($continue =~ m/N|n|No|no|NO/ ) {
		# Save and stop
		$filename = $sectorname . "_CollatedData.csv";
		open (OUTFILE, ">$filename") ||
			die "Unable to open $filename: $!\n";
		print OUTFILE "$starti\n";
		for (my $j=0; $j<=$#systems; $j++) {
			print OUTFILE "$systems[$j][0],$systems[$j][1],$systems[$j][2],$systems[$j][3],$systems[$j][4],";
			print OUTFILE "$systems[$j][5],$systems[$j][6],$systems[$j][7],$systems[$j][8],$systems[$j][9],$systems[$j][10]\n";
		}
		close (OUTFILE);
		$filename = $sectorname . "_Routes.csv";
		open (OUTFILE, ">$filename") ||
			die "Unable to open $filename: $!\n";
		for (my $j=0; $j<$#routes; $j++) {
			print OUTFILE "$routes[$j][0],$routes[$j][1],$routes[$j][2],$routes[$j][3],$routes[$j][4]\n";
		}
		close (OUTFILE);
		exit;
	}
}

# If not, search for Trade Routes
for (my $i = $starti; $i <= $#MajorSystems; $i++) { 
	$source = $NameToID{$MajorSystems[$i]};  
	for (my $destinationI = $i+1; $destinationI <= $#MajorSystems; $destinationI++){ 
		$destination = $NameToID{$MajorSystems[$destinationI]};

		%visited = ();

		@results = breadth2search ( $space, $source, $destination );

		for (my $i = 0; $i < $#results; $i++) {
			SaveEdge ($results[$i], $results[$i+1]);
		}
	}

	# Save the traffic data to the Routes array.
	my @E = $space->edges;
	for (my $k = 0; $k <= $#E; $k++) {
		$source = $E[$k]->[0];
		$destination = $E[$k]->[1];
		$thistraffic = $space->get_edge_attribute( $source, $destination, 'traffic' );
		if ($thistraffic > 0) {
			for (my $r = 0; $r < $#routes; $r++) {
				if (($source eq $routes[$r][1]) && ($destination eq $routes[$r][2])) {
					$found = 1;
					$routes[$r][4] = $thistraffic;
				} elsif (($source eq $routes[$r][2]) && ($destination eq $routes[$r][1])) {
					$found = 1;
					$routes[$r][4] = $thistraffic;
				}
			}
			if (!$found) {
				print "ERROR: Could not find Routes entry for $source to $destination with traffic $thistraffic\n";
			}
			$found = 0;
		}
		$thistraffic = 0;
	}
	$filename = $sectorname . "_CollatedData.csv";
	open (OUTFILE, ">$filename") ||
		die "Unable to open $filename: $!\n";
	print OUTFILE "$i\n";
	for (my $j=0; $j<=$#systems; $j++) {
		print OUTFILE "$systems[$j][0],$systems[$j][1],$systems[$j][2],$systems[$j][3],$systems[$j][4],";
		print OUTFILE "$systems[$j][5],$systems[$j][6],$systems[$j][7],$systems[$j][8],$systems[$j][9],$systems[$j][10]\n";
	}
	close (OUTFILE);
	$filename = $sectorname . "_Routes.csv";
	open (OUTFILE, ">$filename") ||
		die "Unable to open $filename: $!\n";
	for (my $j=0; $j<$#routes; $j++) {
		print OUTFILE "$routes[$j][0],$routes[$j][1],$routes[$j][2],$routes[$j][3],$routes[$j][4]\n";
	}
	close (OUTFILE);
	print "\tCompleted MSR for $MajorSystems[$i] (CTRL-C OK).\n";
}

$thistime = sprintf ("%.3f", (time - $lasttime)/60);
$lasttime = time;
print "\tdone ($thistime min).\n";  

#******************************************************************************************************
#******************************************************************************************************
# Export route data for Astrosynthesis
print "Stage 6: Exporting Route data...\n";

my $mosttraffic = 0;	# Fixed value from previous run of the rest of the program.

for (my $r = 0; $r < $#routes; $r++) {
	if ($routes[$r][4] > $mosttraffic) {
		$mosttraffic = $routes[$r][4]
	}
}

print "\tHighest Traffic rating is $mosttraffic\n";

$filename = $sectorname . "_Routes_AS.csv";
open SECTORLINKS, ">$filename" ||
	die "Unable to open $filename: $!\n";

# Print out the routes in the sector
for (my $r = 0; $r < $#routes; $r++) {
# Route info:
	#Body Type.  This must be 'Route'
	#Start Body.  This is the name of the body that the route segment starts at.
	#End Body.  This is the name of the body that the route segment ends at.
	#Route Type.  This is the type of route.  For example, 'Trade' or 'Hyperjump'
	#Route Name.  This is the name of this particular route.
	print SECTORLINKS "Route,$systems[$IDpointer{$routes[$r][1]}][1],$systems[$IDpointer{$routes[$r][2]}][1],Trade,Route $routes[$r][0] ($routes[$r][4]),";
	#Route Color.  This is the color of the route line, as stored in the format #RRGGBB (similar to an HTML color format).
	# If traffic > 0, green 00FF00 otherwise blue 0000FF
	#Route Style.  This is the line style of the route line (see notes).
	#	 0 - Solid Line
	#	 1 - Dashed Line
	#	 2 - Dotted Line
	#	 3 - Dash-Dot Line
	#	 4 - Dash-Dot-Dot Line
	# If traffic < 1 use 2 otherwise use 0.
	if ($routes[$r][4] > 0) {
		print SECTORLINKS "#00FF00,0,";
	} else {
		print SECTORLINKS "#0000FF,2,";
	}
	#Route Width.  This is the width, in pixels, of the route line.  Must be an integer.
	# Determine the width by the traffic value.  There should be 5 categories of line width. 1-5
	if ($mosttraffic > 100) {
		if ($routes[$r][4] < ($mosttraffic*.2)) {
			print SECTORLINKS "1\n";
		} elsif ($routes[$r][4] < ($mosttraffic*.4)) {
			print SECTORLINKS "2\n";
		} elsif ($routes[$r][4] < ($mosttraffic*.6)) {
			print SECTORLINKS "3\n";
		} elsif ($routes[$r][4] < ($mosttraffic*.8)) {
			print SECTORLINKS "4\n";
		} else {
			print SECTORLINKS "5\n";
		}
	} else {
		print SECTORLINKS "1\n";
	}
}

close (SECTORLINKS);

$thistime = sprintf ("%.3f", (time - $lasttime)/60);
$lasttime = time;
print "\tdone ($thistime min).\n";  

#******************************************************************************************************
#******************************************************************************************************
# Create IP Pro tables
print "Stage 7:Creating Inspiration Pad Pro file...\n";

# Initalize traffic values
for (my $s = 0; $s < $#systems; $s++) {
	$systems[$s][9] = 0;
}
# Add up traffic values.
for (my $r = 0; $r < $#routes; $r++) {
	$systems[$IDpointer{$routes[$r][1]}][9] += $routes[$r][4];
	$systems[$IDpointer{$routes[$r][2]}][9] += $routes[$r][4];
}

my $iptfilename = $sectorname . "_SystemNames.ipt";
open IPTFILE, ">$iptfilename" ||
	die "Unable to open $iptfilename for output: $!\n";
my $tablename = $sectorname . "_SystemName";
print IPTFILE "Table: $tablename\n";
$tablename = $sectorname . "_MainSystems";
print IPTFILE "5:[\@$tablename] (Main System)\n";
$tablename = $sectorname . "_MinorSystems";
print IPTFILE "2:[\@$tablename] (Minor System)\n";
$tablename = $sectorname . "_OffrouteSystems";
print IPTFILE "[\@$tablename] (Offroute System)\n\n";

$tablename = $sectorname . "_MainSystems";
print IPTFILE "Table: $tablename\n";
for (my $g = 0; $g <= $#systems; $g++) {
	if ($systems[$g][6] eq 1) {
		$thisscore = $systems[$g][9] + $systems[$g][5] + 201;
		if ($thisscore < 0) {
			print "ERROR: Negative economic score! $systems[$g][1]: $thisscore.\n";
		}
		print IPTFILE "$thisscore:$systems[$g][1]\n";
	}
}

$tablename = $sectorname . "_MinorSystems";
print IPTFILE "\n\nTable: $tablename\n";
for (my $g = 0; $g <= $#systems; $g++) {
	if ($systems[$g][6] eq 0) {
		if ($systems[$g][9] > 0) {
			$thisscore = $systems[$g][9] + $systems[$g][5] + 201;
			if ($thisscore < 0) {
				print "ERROR: Negative economic score! $systems[$g][1]: $thisscore.\n";
			}
			print IPTFILE "$thisscore:$systems[$g][1]\n";
		}
	}
}

$tablename = $sectorname . "_OffrouteSystems";
print IPTFILE "\n\nTable: $tablename\n";
for (my $g = 0; $g <= $#systems; $g++) {
	if ($systems[$g][6] eq 0) {
		if ($systems[$g][9] < 1) {
			$thisscore = $systems[$g][9] + $systems[$g][5] + 201;
			if ($thisscore < 0) {
				print "ERROR: Negative economic score! $systems[$g][1]: $thisscore.\n";
			}
			print IPTFILE "$thisscore:$systems[$g][1]\n";
		}
	}
}
close (IPTFILE);

# Save data files.
$outputfilename = $sectorname . "_CollatedData.csv";
open (DATAFILE, ">$outputfilename") ||
	die "Unable to open $outputfilename: $!\n";
for (my $j=0; $j<=$#systems; $j++) {
	print DATAFILE "$systems[$j][0],$systems[$j][1],$systems[$j][2],$systems[$j][3],$systems[$j][4],";
	print DATAFILE "$systems[$j][5],$systems[$j][6],$systems[$j][7],$systems[$j][8],$systems[$j][9],$systems[$j][10]\n";
}
close (DATAFILE);

$outputfilename = $sectorname . "_Routes.csv";
open (DATAFILE, ">$outputfilename") ||
	die "Unable to open $outputfilename: $!\n";
for (my $j=0; $j<$#routes; $j++) {
	print DATAFILE "$routes[$j][0],$routes[$j][1],$routes[$j][2],$routes[$j][3],$routes[$j][4]\n";
}
close (DATAFILE);

$thistime = sprintf ("%.3f", (time - $lasttime)/60);
$lasttime = time;
print "\tdone ($thistime min).\n";  

$endtime = time;
$endtime = ($endtime - $starttime)/60;
print "The script took $endtime minutes to run.\n";
